import numpy as np
from .base_k import _BaseK

class KMedians(_BaseK):
    """
    K-Medians clustering object.

    Args:
        n_clusters (int): The number of clusters to form
        dist (str, optional): The distance metric used {'euclidean','manhattan','minikowski','cosine','hamming'}. Defaults to 'manhattan'.
        mini_ord (int, optional): The order for minikowski distance. Ignored if dist is not 'minikowski'. Defaults to 3.
        init (str, optional): The initialisation method used to select initial centroids, 'kmeans++' or 'rand' . Defaults to 'kmeans++'.
        n_init (int, optional): The number of different seeds and times to run the kmeans++ initialisation of which best result is selected. Defaults to 10.
        max_iter (int, optional): The maximum number of iterations the K-Means algorithm can run without convergence. Defaults to 300.
        tol (float, optional): The relative tolerance of the Frobenius norm of the difference in the cluster centers of two consecutive iterations to declare convergence. Defaults to 0.0001.
        random_state (int, optional): The RNG seed used for centroid initialisation. Defaults to 0.
        verb (bool, optional): Enable or disable verbose output. Defaults to False.
        
    Methods:
        fit()
    
    Getter methods call-able after calling fit():
        init_cluster_centers_
        init_labels_
        init_sse_
        cluster_centers_
        labels_
        sse_
        n_iter_converge_
        x_samples_
        x_features_
    """
    def __init__(self, n_clusters, dist='manhattan', mini_ord=3, init='kmeans++', n_init=10, max_iter=300, tol=0.0001, random_state=0, verb=False):
        
        super().__init__(n_clusters, dist, mini_ord, init, n_init, max_iter, tol, random_state, verb)
        self._n_samples = None
        self._n_features = None
        self._init_cluster_centers = None
        self._init_labels = None
        self._init_sse = None
        self._cluster_centers = None
        self._labels = None
        self._sse = None
        self._ssr = None
        self._n_iter_converge = None
        if self._verb:
            print("KMedians object initialised with %s distance metric"%(self._dist))

    def _centroid_new_pos(self, points):
        """
        Takes in a 2D array of all the points belonging to a centroid, returns a vector that is the mean of all points.
        Mean for KMeans, Median for KMedians
        """
        return np.median(points, axis=0)
    
    def _centroids_update(self, n_centroids, X, x_nearest_centroids, current_centroids):
        """
        Returns a vector of the updated centroids.
        x_nearest_centroids denotes the index of the nearest centroid in centroids for each data point.
        This function changes the parameter current_centroid (ndarray) because it is being passed by reference
        
        Args:
            n_centroids (int): the number of centroids
            X (2-dimension ndarray): data
            x_nearest_centroids (1-dimension ndarray): index positions of nearest centroids for each data point
            current_centroids (2-dimension ndarray): the array containing the current positions of the centroids
        """
        current_centroids_copy = current_centroids.copy()
        for i in range(n_centroids):
            centroid_connected_points = X[x_nearest_centroids==i]
            new_centroid = self._centroid_new_pos(centroid_connected_points)
            current_centroids_copy[i] = new_centroid
            
        return current_centroids_copy
    
    def _get_nearest_centroids(self, X, centroids):
        """
        Returns a vector of the index of the closest centroid to each data point in x. Ndarray of ints

        Args:
            X (2-dimension ndarray): data
            centroids (2-dimension ndarra): the array containing the positions of the centroids
        """
        nearest_centroids = np.empty(0, int)
        # Loop over data points
        # Loop over centroids, get closest distance, get index position
        # Add index position to nearest_centroids
        for i in range(X.shape[0]):
            datapoint = X[i]
            distances = []
            for c in centroids:
                distances.append(super()._distance(datapoint, c))
            distances = np.array(distances)
            ind_nearest_centroid = np.argmin(distances)
            nearest_centroids = np.append(nearest_centroids, np.array([ind_nearest_centroid]), 0)

        return nearest_centroids
    
    def fit(self, X):
        """
        Compute k-medians clustering

        Args:
            X (2-dimension ndarray): The X data as a matrix.
        """
        X = self._validate_data(X)
        self._n_samples, self._n_features = X.shape  
        if self._init == "rand":
            initial_centroids = super()._init_random(X)
        elif self._init == "kmeans++":
            initial_centroids = super()._init_kmeansplusplus(X, self._n_clusters)
        else:
            raise ValueError("The init variable %s is invalid " % (self._init))
        if self._verb:
            print("Initial centroid via %s successful" % (self._init), "\n")
        # These variables are for testing the sse of the initial centroids
        initial_labels = self._get_nearest_centroids(X, initial_centroids)
        initial_sse = super()._sse_error(X, initial_centroids, initial_labels)
        self._init_cluster_centers = initial_centroids
        self._init_labels = initial_labels
        self._init_sse = initial_sse
        # Loop over the max_iterations
        current_centroids = np.copy(initial_centroids)
        # 1: Calculate positions of new centroids based on median of points that belong to it
        # 2: Update current_centroids to new_centroids
        if self._verb:
            print('Starting KMedians iterations...')
        for i in range(self._max_iter):
            nearest_centroids = self._get_nearest_centroids(X, current_centroids)
            new_centroids = self._centroids_update(self._n_clusters, X, nearest_centroids, current_centroids)
            # 3: Calculate frobenius norm against tolerance to declare convergence
            diff = new_centroids - current_centroids
            fn_update_diff = np.sqrt(np.sum(np.square(diff)))
            print(fn_update_diff)
            if (fn_update_diff < self._tol):
                if self._verb:
                    print(f"Convergence reached at iteration {i}")
                break
            else:
                current_centroids = new_centroids
        self._n_iter_converge = i + 1
        if self._verb:
            print("KMedians iterations complete...")
        self._cluster_centers = current_centroids
        self._labels = self._get_nearest_centroids(X, self._cluster_centers)
        self._sse = super()._sse_error(X, self._cluster_centers, self._labels)
        self._ssr = super()._sse_residual(X, self._cluster_centers, self._labels)
        
        return self
        
    def fit_transform(self, X):
        """
        Transform X into a cluster-distance matrix. Columns represent cluster centers, rows represent datapoints. 
        Each dimension represents the distance of the datapoint to the cluster center generated after fitting the data with fit() based on the distance metric during intialisation

        Args:
            X (2-dimension ndarray): The X data as a matrix.

        Returns:
            (2-dimension ndarray): cluster-distance matrix
        """

        self.fit(X)
        distance_to_cluster = [[self._distance(i, j) for i in self._cluster_centers] for j in X]
        distance_to_cluster = np.array(distance_to_cluster)
        
        return distance_to_cluster
    
    @property
    def init_cluster_centers_(self):
        return self._init_cluster_centers
    
    @property
    def init_labels_(self):
        return self._init_labels
    
    @property
    def init_sse_(self):
        return self._init_sse
        
    @property
    def cluster_centers_(self):
        return self._cluster_centers
    
    @property
    def labels_(self):
        return self._labels
    
    @property
    def sse_(self):
        return self._sse
    
    @property
    def ssr_(self):
        return self._ssr
    
    @property
    def n_iter_converge(self):
        return self._n_iter_converge
    
    @property
    def x_samples_(self):
        return self._n_samples
    
    @property
    def x_features_(self):
        return self._n_features
    

    


"""
This is an implementation of K-means/medians/medoids with various distance metrics (Euclidean, Manhattan, Cosine...) built over Numpy. 
Read more at https://github.com/jerrold110/Library-Kmars

KMeans parameters:
KMeans(n_clusters, dist='euclidean', mini_ord=3, init='kmeans++', n_init=10, max_iter=300, tol=0.0001, random_state=0, verb=False)
"""
from .kmeans import KMeans
from .kmedians import KMedians
from .kmedoids import KMedoids



import numpy as np
from .modules.distance_metrics import _distance_euclidean, _distance_manhattan, _distance_minikowski, _distance_cosine, _distance_hamming

class _BaseK:
    """
    Base kmeans class. Assumed to be working with numpy arrays
    """
    def __init__(self, n_clusters, dist, mini_ord, init, n_init, max_iter, tol, random_state, verb):
        if dist not in ['euclidean','manhattan','minikowski','cosine','hamming']:
            raise ValueError("dist argument %s not valid" % (dist))
        if init not in ['kmeans++','rand']:
            raise ValueError("init argument %s not valid" % (init))
        
        self._n_clusters = n_clusters
        self._dist = dist
        self._mini_ord = mini_ord
        self._init = init
        self._n_init = n_init
        self._max_iter = max_iter
        self._tol = tol
        self._verb = verb
        self._random_state = random_state
        
    def _validate_data(self, X):
        """
        Validate data, convert into float32 upon calling to standardise the data type of the calculations and the output.
        """
        try:
            if not isinstance(X, np.ndarray):
                raise TypeError
            if len(X.shape)!=2 or X.shape[0]==0 or X.shape[1]==0 or X.dtype =='bool':
                raise ValueError
            return X.astype(np.float32)
        
        except TypeError:
            print("The input must be a numpy array")
        except ValueError:
            print("The array must be 2 dimensional, numerical, and no 0 length arrays")
        except Exception as e:
            print("An error occurred:", e)
        
    def _distance(self, v1, v2):
        """
        Returns vector distance between two points based on distance metric during initialisation.
        Distance calculations are scaled up to float64 to prevent numerical overflow, but returned in float32 to maintain consistency.
        """
        if self._dist == 'euclidean':
            distance = _distance_euclidean(v1, v2)
        elif self._dist == 'manhattan':
            distance = _distance_manhattan(v1, v2)
        elif self._dist == 'minikowski':
            distance = _distance_minikowski(v1, v2, self._mini_ord)
        elif self._dist == 'cosine':
            distance = _distance_cosine(v1, v2)
        elif self._dist == 'hamming':
            distance = _distance_hamming(v1, v2)
        else:
            raise ValueError("We should not be able to get here. Error in _distance function with dist input %s" % (self._dist))

        return distance
    
    def _residual(self, v1, v2):
        """
        Returns residuals of two vectors
        """
        v1 = v1.astype(np.float32)
        v2 = v2.astype(np.float32)
        residual = np.sum(np.abs(v1 - v2))
        
        return residual
    
    def _sse_error(self, X, cluster_centers, x_labels):
        """
        Calculates the sum of squared distances of samples to their closest cluster center based on distance metric during initialisation.

        Args:
            X (2-dimension ndarray): X data
            cluster_centers (2-dimension ndarray): array of vectors, shape[0]=_n_cluster_centers, shape[1] = n_features
            x_labels (1-dimension ndarray): array of ints representing nearest cluster, shape[0] = n_samples

        Returns:
            float: The SSE error
        """
        sse_error = .0
        for i in range(X.shape[0]):
            nearest_cluster_ind = x_labels[i]
            point_to_centroid_distance = self._distance(X[i], cluster_centers[nearest_cluster_ind])
            se_error = point_to_centroid_distance**2
            sse_error += se_error
            
        return sse_error
    
    def _sse_residual(self, X, cluster_centers, x_labels):
        """
        Calculates the sum of squared residuals of samples to their closest cluster center based on distance metric during initialisation.

        Args:
            X (2-dimension ndarray): X data
            cluster_centers (2-dimension ndarray): array of vectors, shape[0]=_n_cluster_centers, shape[1] = n_features
            x_labels (1-dimension ndarray): array of ints representing nearest cluster, shape[0] = n_samples

        Returns:
            float: The SSE error
        """
        sser_residual = .0
        for i in range(X.shape[0]):
            nearest_cluster_ind = x_labels[i]
            point_to_centroid_distance = self._residual(X[i], cluster_centers[nearest_cluster_ind])
            ser_residual = point_to_centroid_distance**2
            sser_residual += ser_residual
            
        return sser_residual
    
    def _init_random(self, X):
        """
        Returns a 2D array of initial centroid positions initialised randomly between the max and min value of every column

        Returns:
            2-dimension ndarray: random centroids
        """
        np.random.seed(self._random_state)
        n_features = X.shape[1]
        random_centroids = np.empty((0,n_features))
        column_maxs = X.max(axis=0)
        column_mins = X.min(axis=0)
        column_ranges = column_maxs - column_mins
        for _ in range(self._n_clusters):
            centroid =  np.random.uniform(0, 1, X.shape[1])
            centroid = column_mins + (centroid * column_ranges)
            random_centroids = np.append(arr=random_centroids, values=np.array([centroid]), axis=0)
            
        return self._validate_data(random_centroids)
    
    def _init_kmeansplusplus(self, X, start_seed):
        """
        Returns an array of initial centroid positions initialised with kmeans++. Based on best SSE with multiple seeds
        Kmeans++ seeks to spread out the k initial clusters.

        Args:
            X (2-dimension ndarray): X data
            start_seed (int): The first RNG seed to be used

        Returns:
            (2-dimension ndarray): array of vectors, shape[0]=_n_cluster_centers, shape[1] = n_features
        """
        # This array stores tuples (see, centroids)
        results = []
        for i in range(self._n_init):
            seed = start_seed + i
            # Initialise the first centroid with random uniform distribution. 
            # Centroids is the array that will be returned
            np.random.seed(seed)
            n_samples, n_features = X.shape
            # Initialise the first centroid with uniform distribution
            centroids = np.empty((0, n_features), np.float32)
            candidate_centroids = X.copy()
            first_centroid_ind = np.random.choice(n_samples)
            first_centroid = X[first_centroid_ind]
            # add new centroid to centroids
            centroids = np.append(arr=centroids, values=np.array([first_centroid]), axis=0)
            # remove datapoint from candidate_centroids
            candidate_centroids = np.delete(arr=candidate_centroids, obj=first_centroid_ind, axis=0)
            # Loop over remaining centroids
            # Loop over all candidate centroids
            # Loop over all centroids. Calculate the min squared distance of all candidate centroids to existing centroids
            # Create the probability distribution and select the next centroid
            np.random.seed(seed)
            for _ in range(self._n_clusters - 1):
                distances = []
                n_candidate_centroids = candidate_centroids.shape[0]
                n_centroids = centroids.shape[0]
                for j in range(n_candidate_centroids):
                    point_to_centroid = []
                    for c in range(n_centroids):
                        d = self._distance(candidate_centroids[j], centroids[c]).astype(np.float64)
                        point_to_centroid.append(d)
                    distances.append(min(point_to_centroid))
                distances = np.array(distances, dtype=np.float32)
                probabilities = distances / np.sum(distances)
                
                next_centroid_ind = np.random.choice(a=n_candidate_centroids, p=probabilities)
                
                # add new centroid to centroids. T
                centroids = np.append(centroids, np.expand_dims(candidate_centroids[next_centroid_ind], 0), 0)
                # remove datapoint from candidate_centroids
                candidate_centroids = np.delete(candidate_centroids, next_centroid_ind, 0)
            # Calculate SSE of the selected candidate centroids
            nearest_centroids = self._get_nearest_centroids(X, centroids)
            sse = self._sse_error(X, centroids, nearest_centroids)
            results.append((sse, centroids))
        # Sort results by sse
        results.sort(key=lambda x:x[0])
        if self._verb:
            print("Kmeans++ initial centroids:")
            print(results[0][1])
            print(results[0][0])

        return self._validate_data(results[0][1])


import numpy as np
from .base_k import _BaseK

class KMedoids(_BaseK):
    """
    K-Medoids clustering object.
    Maximum iterations is set at 50 by default because of the longer complexity of Kmedoids. 
    Convergence is reached if none of the medoids move in a single iteration of n medoids.

    Args:
        n_clusters (int): The number of clusters to form
        dist (str, optional): The distance metric used {'euclidean','manhattan','minikowski','cosine','hamming'}. Defaults to 'euclidean'.
        mini_ord (int, optional): The order for minikowski distance. Ignored if dist is not 'minikowski'. Defaults to 3.
        m_update(str, optional): The type {'se', 'sse'} of metric to decide on which is the best datapoint to swap the medoid and whether to update the medoid
        init (str, optional): The initialisation method used to select initial centroids, 'kmeans++' or 'rand' . Defaults to 'kmeans++'.
        n_init (int, optional): The number of different seeds and times to run the kmeans++ initialisation of which best result is selected. Defaults to 10.
        max_iter (int, optional): The maximum number of iterations the K-Means algorithm can run without convergence. Defaults to 300.
        tol (float, optional): The relative tolerance of the Frobenius norm of the difference in the cluster centers of two consecutive iterations to declare convergence. Defaults to 0.0001.
        random_state (int, optional): The RNG seed used for centroid initialisation. Defaults to 0.
        verb (bool, optional): Enable or disable verbose output. Defaults to False.
        
    Methods:
        fit()
    
    Getter methods call-able after calling fit():
        init_cluster_centers_
        init_labels_
        init_sse_
        cluster_centers_
        labels_
        sse_
        n_iter_converge_
        x_samples
        x_features
    """
    def __init__(self, n_clusters, dist='manhattan', mini_ord=3, m_update='se', init='kmeans++', n_init=10, max_iter=50, tol=0.0001, random_state=0, verb=False):
        
        super().__init__(n_clusters, dist, mini_ord, init, n_init, max_iter, tol, random_state, verb)
        self._fit_score = m_update
        self._n_samples = None
        self._n_features = None
        self._init_cluster_centers = None
        self._init_labels = None
        self._init_sse = None
        self._cluster_centers = None
        self._labels = None
        self._sse = None
        self._ssr = None
        self._n_iter_converge = None
        if self._verb:
            print("KMedoids object initialised with %s distance metric and %s fit score "%(self._dist, self._fit_score))
    
    def _se_error(self, X, cluster_centers, x_labels):
        """
        Calculates the sum of distances of samples to their closest cluster center based on distance metric during initialisation.

        Args:
            X (2-dimension ndarray): X data
            cluster_centers (2-dimension ndarray): array of vectors, shape[0]=_n_cluster_centers, shape[1] = n_features
            x_labels (1-dimension ndarray): array of ints representing nearest cluster, shape[0] = n_samples

        Returns:
            float: The SE error
        """
        se_error = .0
        for i in range(X.shape[0]):
            nearest_cluster_ind = x_labels[i]
            point_to_centroid_distance = self._distance(X[i], cluster_centers[nearest_cluster_ind])
            se_error += point_to_centroid_distance
        return se_error
    
    def _cost(self, X, cluster_centers, x_labels):
        if self._fit_score == 'se':
            return self._se_error(X, cluster_centers, x_labels)
        elif self._fit_score == 'sse':
            return self._sse_error(X, cluster_centers, x_labels)
    
    def _get_nearest_centroids(self, X, centroids):
        """
        Returns a vector of the index of the closest centroid to each data point in x. Ndarray of ints

        Args:
            X (2-dimension ndarray): data
            centroids (2-dimension ndarra): the array containing the positions of the centroids
        """
        nearest_centroids = np.empty(0, int)
        # Loop over data points
        # Loop over centroids, get closest distance, get index position
        # Add index position to nearest_centroids
        for i in range(X.shape[0]):
            datapoint = X[i]
            distances = []
            for c in centroids:
                distances.append(super()._distance(datapoint, c))
            distances = np.array(distances)
            ind_nearest_centroid = np.argmin(distances)
            nearest_centroids = np.append(nearest_centroids, np.array([ind_nearest_centroid]), 0)

        return nearest_centroids
    
    def cluster_centroid_update(self, reference_centroid, cluster_data):
        """
        Selects the best medoid in the cluster to update the medoid to

        Args:
            reference_centroid (1-dimension ndarray): The starting centroid centroid
            cluster_data (2-dimension ndarray): data

        Returns:
            return_centroid: The best centroid to update the data to
        """
        current_cluster_fit_score = self._cost(cluster_data, np.array([reference_centroid]), np.zeros(cluster_data.shape[0], dtype=int))
        best_cluster_fit_score = current_cluster_fit_score
        return_centroid = reference_centroid.copy()
        for new_centroid in cluster_data:
            new_cluster_fit_score = self._cost(cluster_data, np.array([new_centroid]), np.zeros(cluster_data.shape[0], dtype=int))
            if new_cluster_fit_score < best_cluster_fit_score:
                return_centroid = new_centroid
                best_cluster_fit_score = new_cluster_fit_score
                
        return return_centroid
    
    def fit(self, X):
        """
        Compute k-medoids clustering

        Args:
            X (2-dimension ndarray): The X data as a matrix.
        """
        X = self._validate_data(X)
        self._n_samples, self._n_features = X.shape  
        if self._init == "rand":
            initial_centroids = super()._init_random(X)
        elif self._init == "kmeans++":
            initial_centroids = super()._init_kmeansplusplus(X, self._n_clusters)
        else:
            raise ValueError("The init variable %s is invalid " % (self._init))
        if self._verb:
            print("Initial centroid via %s successful" % (self._init), "\n")
        # These variables are for testing the sse of the initial centroids
        initial_labels = self._get_nearest_centroids(X, initial_centroids)
        initial_sse = super()._sse_error(X, initial_centroids, initial_labels)
        self._init_cluster_centers = initial_centroids
        self._init_labels = initial_labels
        self._init_sse = initial_sse
        # Loop over the max_iterations
        current_centroids = np.copy(initial_centroids)
        # assign current_fit_score
        current_fit_score = self._cost(X, initial_centroids, initial_labels)
        # 1: For each iteration, for each cluster
        # 2: Check against all data points, change the medoid in that cluster with that data point
        if self._verb:
            print('Starting KMedoids iterations...')
        for i in range(self._max_iter):
            if self._verb:
                print('Current fit score: %s' % (current_fit_score))
            iteration_centroids = current_centroids
            for n in range(self._n_clusters):
                nearest_centroids = self._get_nearest_centroids(X, current_centroids)
                new_centroids = iteration_centroids.copy()
                new_centroid_n = self.cluster_centroid_update(current_centroids[n], X[nearest_centroids==n])
                new_centroids[n] = new_centroid_n
                new_labels = self._get_nearest_centroids(X, new_centroids)
                new_fit_score = self._cost(X, new_centroids, new_labels)
                #  3: If an iteration improves overall fit score, update the medoid
                if new_fit_score < current_fit_score:
                    current_centroids = new_centroids
                    current_fit_score = new_fit_score
            self._n_iter_converge = i + 1
            # 4: Converge if no change in centroids for an iteration
            if np.array_equal(new_centroids, iteration_centroids):
                if self._verb:
                    print(f"Convergence reached at iteration {i}")
                break
        if self._verb:
            print("KMedoids iterations complete...")
        self._cluster_centers = current_centroids
        self._labels = self._get_nearest_centroids(X, self._cluster_centers)
        self._sse = super()._sse_error(X, self._cluster_centers, self._labels)
        self._ssr = super()._sse_residual(X, self._cluster_centers, self._labels)

        return self
        
    def fit_transform(self, X):
        """
        Transform X into a cluster-distance matrix. Columns represent cluster centers, rows represent datapoints. 
        Each dimension represents the distance of the datapoint to the cluster center generated after fitting the data with fit() based on the distance metric during intialisation

        Args:
            X (2-dimension ndarray): The X data as a matrix.

        Returns:
            (2-dimension ndarray): cluster-distance matrix
        """

        self.fit(X)
        distance_to_cluster = [[self._distance(i, j) for i in self._cluster_centers] for j in X]
        distance_to_cluster = np.array(distance_to_cluster)
        
        return distance_to_cluster
    
    @property
    def init_cluster_centers_(self):
        return self._init_cluster_centers
    
    @property
    def init_labels_(self):
        return self._init_labels
    
    @property
    def init_sse_(self):
        return self._init_sse
        
    @property
    def cluster_centers_(self):
        return self._cluster_centers
    
    @property
    def labels_(self):
        return self._labels
    
    @property
    def sse_(self):
        return self._sse
    
    @property
    def ssr_(self):
        return self._ssr
    
    @property
    def n_iter_converge(self):
        return self._n_iter_converge
    
    @property
    def x_samples_(self):
        return self._n_samples
    
    @property
    def x_features_(self):
        return self._n_features
    

    